"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import shutil
import sys
from distutils.dir_util import remove_tree, copy_tree
from distutils.errors import DistutilsFileError
from logging import Logger
from typing import List

from Setter.Common.ServiceManager import ServiceManager
from Setter.Enums.Software import SoftwareBackup, FactoryResetSoftware, SystemdServices
from Setter.Exceptions.CommandException import CommandException
from Setter.Exceptions.IOErrorException import IOErrorException
from Setter.Logging import Logging
from Setter.WebappConfig import WebappConfig


class FactoryReset:
	"""Factory reset class."""

	_backup_dir: str = '/usr/share/iqrf-gateway/default'
	_logger: Logger = Logging.get_logger()

	@classmethod
	def perform_reset(cls, software: List[FactoryResetSoftware], reset_all: bool) -> None:
		"""
		Perform factory reset for selected software of for all.

		Parameters
		----------
		software: List[FactoryResetSoftware]
			Software configuration to factory reset
		reset_all: bool
			Reset all software managed by backup

		Returns
		-------
		None
		"""
		services_to_reset = []
		# reset cloud provisioning
		if FactoryResetSoftware.CLOUD_PROVISIONING in software or reset_all:
			FactoryReset.reset_cloud_provisioning()
			services_to_reset.append(SystemdServices.CLOUD_PROVISIONING)
		# reset daemon
		if FactoryResetSoftware.DAEMON in software or reset_all:
			FactoryReset.reset_daemon()
			services_to_reset.append(SystemdServices.DAEMON)
		# reset iqaros
		if FactoryResetSoftware.IQAROS in software or reset_all:
			FactoryReset.reset_iqaros()
		# reset network manager
		if FactoryResetSoftware.NETWORK_MANAGER in software or reset_all:
			FactoryReset.reset_network_manager()
			services_to_reset.append(SystemdServices.NETWORK_MANAGER)
		# reset webapp
		if FactoryResetSoftware.WEBAPP in software or reset_all:
			FactoryReset.reset_webapp()
		# restart services
		for service in services_to_reset:
			if ServiceManager.exists(service):
				ServiceManager.restart(service)

	@classmethod
	def reset_cloud_provisioning(cls):
		"""
		Perform Cloud Provisioning factory reset.

		Returns
		-------
		None
		"""
		cls._logger.info('Performing IQRF Cloud Provisioning factory reset.')
		try:
			if sys.version_info >= (3, 8):
				if os.path.exists('/etc/iqrf-cloud-provisioning'):
					shutil.rmtree(path='/etc/iqrf-cloud-provisioning')
				os.makedirs(name='/etc/iqrf-cloud-provisioning', exist_ok=True)
				shutil.copytree(
					src=f'{cls._backup_dir}/{SoftwareBackup.CLOUD_PROVISIONING}/config/',
					dst='/etc/iqrf-cloud-provisioning',
					symlinks=True,
					dirs_exist_ok=True
				)
				if os.path.exists('/usr/share/iqrf-cloud-provisioning/cert'):
					shutil.rmtree(path='/usr/share/iqrf-cloud-provisioning/cert')
				os.makedirs(name='/usr/share/iqrf-cloud-provisioning/cert', exist_ok=True)
				shutil.copytree(
					src=f'{cls._backup_dir}/{SoftwareBackup.CLOUD_PROVISIONING}/certs/',
					dst='/usr/share/iqrf-cloud-provisioning/cert',
					symlinks=True,
					dirs_exist_ok=True
				)
			else:
				if os.path.exists('/etc/iqrf-cloud-provisioning'):
					remove_tree(directory='/etc/iqrf-cloud-provisioning')
				os.makedirs(name='/etc/iqrf-cloud-provisioning', exist_ok=True)
				copy_tree(
					src=f'{cls._backup_dir}/{SoftwareBackup.CLOUD_PROVISIONING}/config/',
					dst='/etc/iqrf-cloud-provisioning',
					preserve_symlinks=True,
					update=1
				)
				if os.path.exists('/usr/share/iqrf-cloud-provisioning/cert'):
					remove_tree(directory='/usr/share/iqrf-cloud-provisioning/cert')
				os.makedirs(name='/usr/share/iqrf-cloud-provisioning/cert/bootstrap', exist_ok=True)
				copy_tree(
					src=f'{cls._backup_dir}/{SoftwareBackup.CLOUD_PROVISIONING}/certs/',
					dst='/usr/share/iqrf-cloud-provisioning/cert',
					preserve_symlinks=True,
					update=1
				)
		except (shutil.Error, DistutilsFileError) as error:
			cls._logger.error(str(error))
			raise IOErrorException(str(error)) from error
		cls._logger.info('IQRF Cloud Provisioning factory reset successful.')

	@classmethod
	def reset_daemon(cls):
		"""
		Perform Daemon factory reset.

		Returns
		-------
		None
		"""
		cls._logger.info('Performing IQRF Gateway Daemon factory reset.')
		try:
			if sys.version_info >= (3, 8):
				os.makedirs(name='/etc/iqrf-gateway-daemon', exist_ok=True)
				shutil.copytree(
					src=f'{cls._backup_dir}/{SoftwareBackup.DAEMON}/config/',
					dst='/etc/iqrf-gateway-daemon',
					symlinks=True,
					dirs_exist_ok=True
				)
				if os.path.exists('/usr/share/iqrf-gateway-daemon/cfgSchemas'):
					shutil.rmtree(path='/usr/share/iqrf-gateway-daemon/cfgSchemas')
				os.makedirs(name='/usr/share/iqrf-gateway-daemon/cfgSchemas')
				shutil.copytree(
					src=f'{cls._backup_dir}/{SoftwareBackup.DAEMON}/cfgSchemas/',
					dst='/usr/share/iqrf-gateway-daemon/cfgSchemas',
					symlinks=True,
					dirs_exist_ok=True
				)
				if os.path.exists('/usr/share/iqrf-gateway-daemon/schedulerSchemas'):
					shutil.rmtree(path='/usr/share/iqrf-gateway-daemon/schedulerSchemas')
				os.makedirs(name='/usr/share/iqrf-gateway-daemon/schedulerSchemas')
				shutil.copytree(
					src=f'{cls._backup_dir}/{SoftwareBackup.DAEMON}/schedulerSchemas/',
					dst='/usr/share/iqrf-gateway-daemon/schedulerSchemas',
					symlinks=True,
					dirs_exist_ok=True
				)
				if os.path.exists('/var/cache/iqrf-gateway-daemon/scheduler'):
					shutil.rmtree(path='/var/cache/iqrf-gateway-daemon/scheduler')
				os.makedirs(name='/var/cache/iqrf-gateway-daemon/scheduler')
				shutil.copytree(
					src=f'{cls._backup_dir}/{SoftwareBackup.DAEMON}/scheduler/',
					dst='/var/cache/iqrf-gateway-daemon/scheduler',
					symlinks=True,
					dirs_exist_ok=True
				)
			else:
				os.makedirs(name='/etc/iqrf-gateway-daemon', exist_ok=True)
				copy_tree(
					src=f'{cls._backup_dir}/{SoftwareBackup.DAEMON}/config/',
					dst='/etc/iqrf-gateway-daemon',
					preserve_symlinks=True,
					update=1
				)
				if os.path.exists('/usr/share/iqrf-gateway-daemon/cfgSchemas'):
					remove_tree(directory='/usr/share/iqrf-gateway-daemon/cfgSchemas')
				os.makedirs(name='/usr/share/iqrf-gateway-daemon/cfgSchemas')
				copy_tree(
					src=f'{cls._backup_dir}/{SoftwareBackup.DAEMON}/cfgSchemas/',
					dst='/usr/share/iqrf-gateway-daemon/cfgSchemas',
					preserve_symlinks=True,
					update=1
				)
				if os.path.exists('/usr/share/iqrf-gateway-daemon/schedulerSchemas'):
					remove_tree(directory='/usr/share/iqrf-gateway-daemon/schedulerSchemas')
				os.makedirs(name='/usr/share/iqrf-gateway-daemon/schedulerSchemas')
				copy_tree(
					src=f'{cls._backup_dir}/{SoftwareBackup.DAEMON}/schedulerSchemas/',
					dst='/usr/share/iqrf-gateway-daemon/schedulerSchemas',
					preserve_symlinks=True,
					update=1
				)
				if os.path.exists('/var/cache/iqrf-gateway-daemon/scheduler'):
					remove_tree(directory='/var/cache/iqrf-gateway-daemon/scheduler')
				os.makedirs(name='/var/cache/iqrf-gateway-daemon/scheduler')
				copy_tree(
					src=f'{cls._backup_dir}/{SoftwareBackup.DAEMON}/scheduler/',
					dst='/var/cache/iqrf-gateway-daemon/scheduler',
					preserve_symlinks=True,
					update=1
				)
		except (shutil.Error, DistutilsFileError) as error:
			cls._logger.error(str(error))
			raise IOErrorException(str(error)) from error
		cls._logger.info('IQRF Gateway Daemon factory reset successful.')

	@classmethod
	def reset_iqaros(cls):
		"""
		Perform IQAROS factory reset.

		Returns
		-------
		None
		"""
		cls._logger.info('Performing IQAROS factory reset.')
		result = WebappConfig.run_manager_command('user:remove-all --role basic --no-interaction')
		if result.returncode != 0:
			error_str = f'Failed to remove IQAROS users: {result.stderr}.'
			cls._logger.error(error_str)
			raise CommandException(error_str)
		cls._logger.info('IQAROS factory reset successful.')

	@classmethod
	def reset_network_manager(cls):
		"""
		Perform NetworkManager factory reset.

		Returns
		-------
		None
		"""
		cls._logger.info('Performing NetworkManager factory reset.')
		try:
			if sys.version_info >= (3, 8):
				if os.path.exists('/etc/NetworkManager/system-connections'):
					shutil.rmtree(path='/etc/NetworkManager/system-connections')
				shutil.copytree(
					src=f'{cls._backup_dir}/{SoftwareBackup.NETWORK_MANAGER}',
					dst='/etc/NetworkManager',
					symlinks=True,
					dirs_exist_ok=True
				)
			else:
				if os.path.exists('/etc/NetworkManager/system-connections'):
					remove_tree(directory='/etc/NetworkManager/system-connections')
				copy_tree(
					src=f'{cls._backup_dir}/{SoftwareBackup.NETWORK_MANAGER}',
					dst='/etc/NetworkManager',
					preserve_symlinks=1,
					update=1
				)
			cls._logger.info('NetworkManager factory reset successful.')
		except (shutil.Error, DistutilsFileError) as error:
			cls._logger.error(str(error))
			raise IOErrorException(str(error)) from error

	@classmethod
	def reset_webapp(cls):
		"""
		Perform Webapp factory reset.

		Returns
		-------
		None
		"""
		cls._logger.info('Performing IQRF Gateway Webapp factory reset.')
		if os.path.exists('/var/lib/iqrf-gateway-webapp/database.db'):
			os.remove('/var/lib/iqrf-gateway-webapp/database.db')
		shutil.rmtree(path='/var/cache/iqrf-gateway-webapp/cache/')
		shutil.rmtree(path='/var/cache/iqrf-gateway-webapp/sessions/')
		WebappConfig.configure_database()
		cls._logger.info('IQRF Gateway Webapp factory reset successful.')
